//---------------------------------
// 2019/3/15

#include <Wire.h>
#include "BME280.h"


//---------------------------------
// Initialization

BME280::BME280(){
  addrI2C = 0;
  adcT = 0;
  adcP = 0;
  adcH = 0;
  compT = 0;
  compP = 0;
  compH = 0;
  T = 0;
  P = 0;
  H = 0;

  for(uint8_t i=0; i<calLength; i++){
    cal.byte[i] = 0;
  }
}

bool BME280::begin(uint8_t addrI2C){
  Wire.begin();
  this->addrI2C = addrI2C;
  if(!readID()){
    return false;
  }
  readCal();
  return true;
}

//---------------------------------
// I2C Transmission

// Read 1Byte from register
uint8_t BME280::read(uint8_t addr){
  Wire.beginTransmission(addrI2C);
  Wire.write(addr);
  Wire.endTransmission();
  Wire.requestFrom(addrI2C,1);
  return Wire.read();
}

// Read length Byte from address
void BME280::read(uint8_t addr, uint8_t*data, uint8_t length){
  Wire.beginTransmission(addrI2C);
  Wire.write(addr);
  Wire.endTransmission();
  Wire.requestFrom(addrI2C,length);
  for(uint8_t i=0; i<length; i++){
    data[i] = Wire.read();
  }
}

// Write 1Byte at specific address
void BME280::write(uint8_t addr, uint8_t data){
  Wire.beginTransmission(addrI2C);
  Wire.write(addr);
  Wire.write(data);
  Wire.endTransmission();
  return;
}

// Write length Byte at specific address
void BME280::write(uint8_t addr, uint8_t*data, uint8_t length){
  Wire.beginTransmission(addrI2C);
  Wire.write(addr);
  for(uint8_t i=0; i<length; i++){
      Wire.write(data[i]);
  }
  Wire.endTransmission();
  return;
}

//---------------------------------
// Register Operation

// Read ID
//  Return true if success
bool BME280::readID(){
  uint8_t data = read(0xD0);
  if(data!=0x60){
    return false;
  }
  return true;
}

// Read status register
uint8_t BME280::readStatus(){
  uint8_t data = read(0xF3);
  return data&0x9;
}

// Read calibration
void BME280::readCal(){
  uint8_t buf[2];
  read(0x88, cal.byte, calLengthTP);  // Cal T and P are read sequentially

  // H1
  read(0xA1, buf, 1);
  cal.byte[24] = buf[0];

  // H2
  read(0xE1, buf, 1);
  cal.byte[26] = buf[0];
  read(0xE2, buf, 1);
  cal.byte[27] = buf[0];

  // H3
  read(0xE3, buf, 1);
  cal.byte[28] = buf[0];

  // H4
  read(0xE4, buf, 2);
  cal.byte[30] = ((buf[0]&0xF)<<4)+(buf[1]&0xF);
  cal.byte[31] = buf[0]>>4;

  // H5
  read(0xE5, buf, 2);
  cal.byte[32] = (buf[0]>>4)+((buf[1]&0xF)<<4);
  cal.byte[33] = buf[1]>>4;

  // H6
  read(0xE7, buf, 1);
  cal.byte[34] = buf[0];
}

// Read ADC registers of T, P and H
// Calculated compensate data
void BME280::readADC(){
  uint8_t buf[8];
  read(0xF7, buf, 8);
  adcP = ((uint32_t)buf[0]<<12) + ((uint32_t)buf[1]<<4) + (buf[2]>>4);
  adcT = ((uint32_t)buf[3]<<12) + ((uint32_t)buf[4]<<4) + (buf[5]>>4);
  adcH = ((uint32_t)buf[6]<<8) + buf[7];
  compT = compensate_T_int32(adcT);
  compP = compensate_P_int64(adcP);
  compH = compensate_H_int32(adcH);
  T = ((float)(int32_t)compT/100);
  P = ((float)compP)/25600;
  H = ((float)compH)/1024;
}


//---------------------------------
// Measurement

// Measure forced mode
//  Parameters
//   pooling : If true, wait until measurement completion by read status
void BME280::forced(bool pooling){
  write(0xF5, 0x0);  // config
  write(0xF2, 0x5);  // ctrl_hum, oversampling x16
  write(0xF4, 0xB5);  // ctrl_meas, oversampling x16, forced mode
  if(!pooling){
    return;
  }
  delay(1);
  while(0!=readStatus()){
    delay(1);
  }
}



//---------------------------------
// Print Parameters and Results

// Print Calibration Data
void BME280::printCal(){
  char c[32];
  Serial.print(F("Calibration Data\n"));
  Serial.println(String(F(" dig_T1 : ")) + String(cal.dig_T1));
  Serial.println(String(F(" dig_T2 : ")) + String(cal.dig_T2));
  Serial.println(String(F(" dig_T3 : ")) + String(cal.dig_T3));
  Serial.println(String(F(" dig_P1 : ")) + String(cal.dig_P1));
  Serial.println(String(F(" dig_P2 : ")) + String(cal.dig_P2));
  Serial.println(String(F(" dig_P3 : ")) + String(cal.dig_P3));
  Serial.println(String(F(" dig_P4 : ")) + String(cal.dig_P4));
  Serial.println(String(F(" dig_P5 : ")) + String(cal.dig_P5));
  Serial.println(String(F(" dig_P6 : ")) + String(cal.dig_P6));
  Serial.println(String(F(" dig_P7 : ")) + String(cal.dig_P7));
  Serial.println(String(F(" dig_P8 : ")) + String(cal.dig_P8));
  Serial.println(String(F(" dig_P9 : ")) + String(cal.dig_P9));
  Serial.println(String(F(" dig_H1 : ")) + String(cal.dig_H1));
  Serial.println(String(F(" dig_H2 : ")) + String(cal.dig_H2));
  Serial.println(String(F(" dig_H3 : ")) + String(cal.dig_H3));
  Serial.println(String(F(" dig_H4 : ")) + String(cal.dig_H4));
  Serial.println(String(F(" dig_H5 : ")) + String(cal.dig_H5));
  Serial.println(String(F(" dig_H6 : ")) + String(cal.dig_H6));
}

// Print ADC Register and Compensated Data
void BME280::printRegister(){
  Serial.print(F("Register Data\n"));
  Serial.println(String(F(" ADC T : 0x")) + String(adcT, HEX));
  Serial.println(String(F(" ADC P : 0x")) + String(adcP, HEX));
  Serial.println(String(F(" ADC H : 0x")) + String(adcH, HEX));
  Serial.println(String(F(" Comp T : ")) + String(compT));
  Serial.println(String(F(" Comp P : ")) + String(compP));
  Serial.println(String(F(" Comp H : ")) + String(compH));
}

// Print Measurement Results
void BME280::printMeas(){
  char c[8];
  Serial.print(F("Measured Data\n"));
  Serial.println(String(F(" Temp : ")) + dtostrf(T, 3, 1, c) + String(F("C")));
  Serial.println(String(F(" Pressure : ")) + dtostrf(P, 4, 1, c) + String(F("hPa")));
  Serial.println(String(F(" Humidity : ")) + dtostrf(H, 2, 1, c) + String(F("%")));
}


//---------------------------------
// Calibration Computation

// Returns temperature in DegC, resolution is 0.01 DegC. Output value of "5123" equals 51.23 DegC.
// t_fine carries fine temperature as global value
BME280_S32_t BME280::compensate_T_int32(BME280_S32_t adc_T)
{
  BME280_S32_t var1, var2, T;
  var1  = ((((adc_T>>3) - ((BME280_S32_t)cal.dig_T1<<1))) * ((BME280_S32_t)cal.dig_T2)) >> 11;
  var2  = (((((adc_T>>4) - ((BME280_S32_t)cal.dig_T1)) * ((adc_T>>4) - ((BME280_S32_t)cal.dig_T1))) >> 12) *
    ((BME280_S32_t)cal.dig_T3)) >> 14;
  t_fine = var1 + var2;
  T  = (t_fine * 5 + 128) >> 8;
  return T;
}

// Returns pressure in Pa as unsigned 32 bit integer in Q24.8 format (24 integer bits and 8 fractional bits).
// Output value of "24674867" represents 24674867/256 = 96386.2 Pa = 963.862 hPa
BME280_U32_t BME280::compensate_P_int64(BME280_S32_t adc_P)
{
  BME280_S64_t var1, var2, p;
  var1 = ((BME280_S64_t)t_fine) - 128000;
  var2 = var1 * var1 * (BME280_S64_t)cal.dig_P6;
  var2 = var2 + ((var1*(BME280_S64_t)cal.dig_P5)<<17);
  var2 = var2 + (((BME280_S64_t)cal.dig_P4)<<35);
  var1 = ((var1 * var1 * (BME280_S64_t)cal.dig_P3)>>8) + ((var1 * (BME280_S64_t)cal.dig_P2)<<12);
  var1 = (((((BME280_S64_t)1)<<47)+var1))*((BME280_S64_t)cal.dig_P1)>>33;
  if (var1 == 0)
  {
    return 0; // avoid exception caused by division by zero
  }
  p = 1048576-adc_P;
  p = (((p<<31)-var2)*3125)/var1;
  var1 = (((BME280_S64_t)cal.dig_P9) * (p>>13) * (p>>13)) >> 25;
  var2 = (((BME280_S64_t)cal.dig_P8) * p) >> 19;
  p = ((p + var1 + var2) >> 8) + (((BME280_S64_t)cal.dig_P7)<<4);
  return (BME280_U32_t)p;
}

// Returns humidity in %RH as unsigned 32 bit integer in Q22.10 format (22 integer and 10 fractional bits).
// Output value of "47445" represents 47445/1024 = 46.333 %RH
BME280_U32_t BME280::compensate_H_int32(BME280_S32_t adc_H)
{
  BME280_S32_t v_x1_u32r;
  v_x1_u32r = (t_fine - ((BME280_S32_t)76800));
  v_x1_u32r = (((((adc_H << 14) - (((BME280_S32_t)cal.dig_H4) << 20) - (((BME280_S32_t)cal.dig_H5) * v_x1_u32r)) +
    ((BME280_S32_t)16384)) >> 15) * (((((((v_x1_u32r * ((BME280_S32_t)cal.dig_H6)) >> 10) * (((v_x1_u32r *
    ((BME280_S32_t)cal.dig_H3)) >> 11) + ((BME280_S32_t)32768))) >> 10) + ((BME280_S32_t)2097152)) *
    ((BME280_S32_t)cal.dig_H2) + 8192) >> 14));
  v_x1_u32r = (v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((BME280_S32_t)cal.dig_H1)) >> 4));
  v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
  v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);
  return (BME280_U32_t)(v_x1_u32r>>12); 
}
